﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace Ratchet01
{
	public partial class MainPage : UserControl
	{
		// 画面の大きさ
		const uint DISP_WIDTH = 800;
		const uint DISP_HEIGHT = 400;
		// 原点
		const uint CENTER_X = DISP_WIDTH / 2;
		const uint CENTER_Y = DISP_HEIGHT / 2;

		// 各種パラメータ
		const uint R_BALL = 12;   // ボールの直径
		double X0, Y0;	// ボールの座標
		double Vx;	// ボールの速度 dx/dt

		int Count_Plus = 0, Count_Moinus = 0;
        int Clock;	// 時刻

        // ランダム力パラメータ
		bool CHK_rnd_force = false;
		const double F_RND_FACTOR = 0.000002;
		double F_rnd = 50 * F_RND_FACTOR;   // ランダム力の大きさ

        // Sin波パラメータ
		bool[] CHK_sin_force = {false, false, false};
		const double ALPHA_FACTOR = 0.000005; // 0.000003;
		double[] Alpha = { 0.0, 0.0, 0.0 };	// Sin波の大きさ
		const double THETA_MINI = 0.5;
		const double THETA_FACTOR = 0.5;
		double[] Theta = { THETA_MINI, THETA_MINI, THETA_MINI };	// Sin波の周期
		double[] Phase = { 0.0, 0.0, 0.0 };	// 位相のずれ

        // ラチェットパラメータ
		const double INTERVAL = 0.1;    // １ラチェットの間隔
        double RC_asym = 0.05;      // 頂点の位置（非対称の度合い）
		const double RC_BIAS_FACTOR = (0.00001 / 100);
		double RC_bias = 20 * RC_BIAS_FACTOR;	// ラチェットの高さ

		List<Line> ratchet_lines = new List<Line>();

		// 描画オブジェクト
		MyCircle Ball;	// ボール

		Random Rnd;	// 乱数
		
		public MainPage()
		{
			InitializeComponent();

			initObjects();
			CompositionTarget.Rendering += new EventHandler(MainHandler);
		}

		// 変数の初期化
		private void initVars()
		{
			// 各種パラメータの初期化
			X0 = 0.0;	Y0 = 0.0;
            Vx = initForce();
			Clock = 0;
		}

		// オブジェクトの初期化
		private void initObjects()
		{
			// 乱数の初期化
			Rnd = new System.Random();

			initVars();
			
			// ボールを設置
			Ball = new MyCircle(dispX(X0), dispY(Y0), R_BALL);
			this.LayoutRoot.Children.Add(Ball.obj);

			// 基準線を引く
			drawLine(-1, 0, 1, 0);
			drawLine(0, -1, 0, 1);

			// ラチェットを描く
			drawRatcet();
		}

		const uint RC_BIAS_DISP_FACTOR = 40000;   // ラチェットの高さ表示拡大率

		// ラチェットを描く
		private void drawRatcet()
		{
			int times = (int)(1.0 / INTERVAL);
			for (int i = -times; i < times; i++)
			{
				Line a_line;
				double xx = i * INTERVAL;

				a_line = drawLine(xx, 0, xx + RC_asym, -RC_bias * RC_BIAS_DISP_FACTOR);
				ratchet_lines.Add(a_line);
				a_line = drawLine(xx + RC_asym, -RC_bias * RC_BIAS_DISP_FACTOR, xx + INTERVAL, 0);
				ratchet_lines.Add(a_line);
			}
		}

		// ラチェットを再描画する
		private void redrawRatchet()
		{
			int times = (int)(1.0 / INTERVAL);
			for (int i = 0; i < ratchet_lines.Count; i++)
			{
				Line a_line = ratchet_lines[i];
				double xx;

				if (i % 2 == 0) // 偶数
				{
					xx = (i / 2 - times) * INTERVAL;
					a_line.X2 = dispX(xx + RC_asym);
					a_line.Y2 = dispY(-RC_bias * RC_BIAS_DISP_FACTOR);
				}
				else // 奇数
				{
					xx = ((i-1) / 2 - times) * INTERVAL;
					a_line.X1 = dispX(xx + RC_asym);
					a_line.Y1 = dispY(-RC_bias * RC_BIAS_DISP_FACTOR);
				}
			}
		}

		// 線を描き足す
		private Line drawLine(double x0, double y0, double x1, double y1)
		{
			Line a_line = new Line();
			a_line.X1 = dispX(x0);
			a_line.Y1 = dispY(y0);
			a_line.X2 = dispX(x1);
			a_line.Y2 = dispY(y1);
			a_line.Stroke = new SolidColorBrush(Colors.Black);
			this.LayoutRoot.Children.Add(a_line);

			return a_line;
		}

		// タイマーで呼び出されるメインハンドラー
		private void MainHandler(object o, EventArgs sender)
		{
			// ボールの移動
			updateAction();

            // 終了判定
            checkEnd();

			// ボールの描画
			Ball.x = dispX(X0);
		//	ball.y = dispY(y0);
			Ball.Action();

			Clock++;	// 全体の時刻を進める

			// 数字の表示、動いていると何となくそれっぽい
		//	DebugText.Text
			//	= "(" + x0.ToString("F") + ")";
			//	+", " + y0.ToString("F") + ")";
		}

		// ボールの移動
		private void updateAction()
		{
            // ランダムな力を加えてみる
			Vx += rndFoerce();

			// 周期的な力を加えてみる
            Vx += sinForce(Clock);

			// ラチェット・ポテンシャルを加える
            Vx += rachetForce();

			X0 += Vx;   // ボールの位置を移動
		}

        // ラチェット・ポテンシャル力
        // 力の正負によってボールの色を変えている
        private double rachetForce()
        {
            double retval = 0.0;

            double mod = X0 % INTERVAL;
            if (mod < 0) { mod += INTERVAL; } // 負の%補正、余りを常に正とする

            if (mod <= RC_asym)
            {
                retval = - (RC_bias / Math.Sqrt(RC_asym * RC_asym + RC_bias * RC_bias));

				Ball.setColor(Color.FromArgb(0xFF, 0xFF, 0x99, 0x00)); // オレンジ色
            }
            else
            {
                retval = (RC_bias / Math.Sqrt((INTERVAL - RC_asym) * (INTERVAL - RC_asym) + RC_bias * RC_bias));

				Ball.setColor(Color.FromArgb(0xFF, 0x00, 0xFF, 0x00)); // 緑色
            }

            return retval;
        }

        // ランダムな力
        private double rndFoerce()
        {
            double retval = 0.0;
			
			if (CHK_rnd_force == false) // チェックされていなければ
			{
				return retval;	// 何もせずリターン
			}

            if (Rnd.Next(2) == 0) // 0 or 1
            {
                retval = F_rnd;  // プラス
            }
            else
            {
                retval = -F_rnd; // マイナス
            }
            return retval;
        }

        // 周期的な力
        private double sinForce(double t_in)
        {
            double retval = 0.0;
            double th = deg2rad(t_in);

            // 複数の周期的な力を足し合わせてみる
			for (int i = 0; i < 3; i++)
			{
				if (CHK_sin_force[i] == true)
				{
					retval += Alpha[i] * Math.Sin(Theta[i] * (th + Phase[i]));
				}
			}

            return retval;
        }

        // 周期的な力のVX初期値設定
		// ちょうど中央に落ち着くようなVX初期値を探す
		private double initForce()
		{
			// ここで位相をランダムに設定する
			Phase[0] = deg2rad(Rnd.Next(2 * 360));
			Phase[1] = deg2rad(Rnd.Next(2 * 360));
			Phase[2] = deg2rad(Rnd.Next(2 * 360));

			// 360度 x 10回 = 3600回実行した平均をとる
			double vx = 0.0;
			double xx = 0.0;

			for (int t = 0; t < 3600; t++)
			{
                vx += sinForce(t);
				xx += vx;
			}

			return -(xx / 3600);
		}

        // 終了判定、はみ出したらリセットする
        private void checkEnd()
        {
            if (X0 > 1.0)
            {
				++Count_Plus;
				int total_count = Count_Plus + Count_Moinus;
				output_text.Text += (total_count + ": (+)" + Count_Plus + ", clock:" + Clock + "\n");
                initVars();
            }
			else if (X0 < -1.0)
			{
				++Count_Moinus;
				int total_count = Count_Plus + Count_Moinus;
				output_text.Text += (total_count + ": (-)" + Count_Moinus + ", clock:" + Clock + "\n");
				initVars();
			}
        }

		// 角度からラジアン角への変換
		private double deg2rad(double deg)
		{
			return deg * Math.PI / 180;
		}

		// 画面座標への変換
		private double dispX(double x)
		{
			return (x * CENTER_X + CENTER_X);
		}
		private double dispY(double y)
		{
			return (y * CENTER_Y + CENTER_Y);
		}

		////////////////////////////////////////////////////////////////
		// UIコンポーネント

		// ボタンを押したらリスタート
		private void Do_btn1_Click(object sender, RoutedEventArgs e)
		{
			initVars();

			// テキスト表示をクリアー
			output_text.Text = "";
			Count_Plus = 0;
			Count_Moinus = 0;
		}

		// ラチェットの高さスライドバー
		private void bias_slide_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			if (bias_slide != null)
			{
				RC_bias = RC_BIAS_FACTOR * bias_slide.Value;
				bias_text.Text = bias_slide.Value.ToString("#.0");
			}
			redrawRatchet();	// ラチェットを再描画する
		}

		// ラチェットの偏りスライドバー
		private void asym_slide_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			if (asym_slide != null)
			{
				RC_asym = (INTERVAL / 100) * asym_slide.Value;
			}
			// 両極端は捨てる
			if (RC_asym < 0.01) { RC_asym = 0.01; }
			else if (RC_asym > 0.09) { RC_asym = 0.09; }

			if (asym_text != null)
			{
				asym_text.Text = RC_asym.ToString("#.000");
			}
			redrawRatchet();	// ラチェットを再描画する
		}

		// ランダム力チェック
		private void rnd_check_Checked(object sender, RoutedEventArgs e)
		{
			CHK_rnd_force = (bool)rnd_check.IsChecked;
		}

		// サイン力１チェック
		private void sin1_check_Checked(object sender, RoutedEventArgs e)
		{
			CHK_sin_force[0] = (bool)sin1_check.IsChecked;
		}

		// サイン力２チェック
		private void sin2_check_Checked(object sender, RoutedEventArgs e)
		{
			CHK_sin_force[1] = (bool)sin2_check.IsChecked;
		}

		// サイン力３チェック
		private void sin3_check_Checked(object sender, RoutedEventArgs e)
		{
			CHK_sin_force[2] = (bool)sin3_check.IsChecked;
		}

		// ランダム力スライドバー
		private void frnd_slied_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			if (frnd_slied != null)
			{
				F_rnd = F_RND_FACTOR * frnd_slied.Value;
				frnd_text.Text = frnd_slied.Value.ToString("#.0");
			}
		}

		// サイン力１大きさスライドバー
		private void sin1_al_slide_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			Alpha[0] = ALPHA_FACTOR * sin1_al_slide.Value;
			sin1_al_text.Text = sin1_al_slide.Value.ToString("#.000");
			Vx = initForce();
		}

		// サイン力１周波数スライドバー
		private void sin1_th_slide_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			Theta[0] = THETA_MINI + THETA_FACTOR * sin1_th_slide.Value;
			sin1_th_text.Text = Theta[0].ToString("#.000");
			Vx = initForce();
		}

		// サイン力２大きさスライドバー
		private void sin2_al_slide_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			Alpha[1] = ALPHA_FACTOR * sin2_al_slide.Value;
			sin2_al_text.Text = sin2_al_slide.Value.ToString("#.000");
			Vx = initForce();
		}

		// サイン力２周波数スライドバー
		private void sin2_th_slide_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			Theta[1] = THETA_MINI + THETA_FACTOR * sin2_th_slide.Value;
			sin2_th_text.Text = Theta[1].ToString("#.000");
			Vx = initForce();
		}

		// サイン力３大きさスライドバー
		private void sin3_al_slide_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			Alpha[2] = ALPHA_FACTOR * sin3_al_slide.Value;
			sin3_al_text.Text = sin3_al_slide.Value.ToString("#.000");
			Vx = initForce();
		}

		// サイン力３周波数スライドバー
		private void sin3_th_slide_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
		{
			Theta[2] = THETA_MINI + THETA_FACTOR * sin3_th_slide.Value;
			sin3_th_text.Text = Theta[2].ToString("#.000");
			Vx = initForce();
		}
	}

	/**
	 *  ボールオブジェクト
	 */
	class MyCircle
	{
		public Ellipse obj;
		public double r;
		public double x, y;

		public MyCircle(double x0, double y0, double r0)
		{
			this.obj = new Ellipse();
			this.r = r0;
			this.x = x0;
			this.y = y0;
			this.obj.Width = 2 * r0;
			this.obj.Height = 2 * r0;

			this.setColor(Color.FromArgb(0xFF, 0xFF, 0x99, 0x00)); //オレンジ色
		}

		public void setColor(Color c)
		{
			SolidColorBrush brush = new SolidColorBrush(c);
			this.obj.Fill = brush;
		}

		public void Action()
		{
			this.obj.SetValue(Canvas.LeftProperty, this.x - this.r);
			this.obj.SetValue(Canvas.TopProperty, this.y - this.r);
		}
	}
}
